@@ -2,19 +2,24 @@
 A PostScript backend, which can produce both PostScript .ps and .eps.
 """
 
+import bisect
 import codecs
 import datetime
 from enum import Enum
 import functools
 from io import StringIO
+import itertools
 import logging
+import math
 import os
 import pathlib
-import re
 import shutil
+import struct
 from tempfile import TemporaryDirectory
+import textwrap
 import time
 
+import fontTools
 import numpy as np
 
 import matplotlib as mpl
@@ -24,8 +29,7 @@ from matplotlib.backend_bases import (
     _Backend, FigureCanvasBase, FigureManagerBase, RendererBase)
 from matplotlib.cbook import is_writable_file_like, file_requires_unicode
 from matplotlib.font_manager import get_font
-from matplotlib.ft2font import LOAD_NO_SCALE, FT2Font
-from matplotlib._ttconv import convert_ttf_to_ps
+from matplotlib.ft2font import LoadFlags
 from matplotlib._mathtext_data import uni2type1
 from matplotlib.path import Path
 from matplotlib.texmanager import TexManager
@@ -33,20 +37,11 @@ from matplotlib.transforms import Affine2D
 from matplotlib.backends.backend_mixed import MixedModeRenderer
 from . import _backend_pdf_ps
 
+
 _log = logging.getLogger(__name__)
-
-backend_version = 'Level II'
 debugPS = False
 
 
-class PsBackendHelper:
-    def __init__(self):
-        self._cached = {}
-
-
-ps_backend_helper = PsBackendHelper()
-
-
 papersize = {'letter': (8.5, 11),
              'legal': (8.5, 14),
              'ledger': (11, 17),
@@ -74,31 +69,8 @@ papersize = {'letter': (8.5, 11),
              'b10': (1.26, 1.76)}
 
 
-def _get_papertype(w, h):
-    for key, (pw, ph) in sorted(papersize.items(), reverse=True):
-        if key.startswith('l'):
-            continue
-        if w < pw and h < ph:
-            return key
-    return 'a0'
-
-
-def _nums_to_str(*args):
-    return " ".join(f"{arg:1.3f}".rstrip("0").rstrip(".") for arg in args)
-
-
-@_api.deprecated("3.6", alternative="a vendored copy of this function")
-def quote_ps_string(s):
-    """
-    Quote dangerous characters of S for use in a PostScript string constant.
-    """
-    s = s.replace(b"\\", b"\\\\")
-    s = s.replace(b"(", b"\\(")
-    s = s.replace(b")", b"\\)")
-    s = s.replace(b"'", b"\\251")
-    s = s.replace(b"`", b"\\301")
-    s = re.sub(br"[^ -~\n]", lambda x: br"\%03o" % ord(x.group()), s)
-    return s.decode('ascii')
+def _nums_to_str(*args, sep=" "):
+    return sep.join(f"{arg:1.3f}".rstrip("0").rstrip(".") for arg in args)
 
 
 def _move_path_to_path_or_stream(src, dst):
@@ -108,7 +80,7 @@ def _move_path_to_path_or_stream(src, dst):
     If *dst* is a path, the metadata of *src* are *not* copied.
     """
     if is_writable_file_like(dst):
-        fh = (open(src, 'r', encoding='latin-1')
+        fh = (open(src, encoding='latin-1')
               if file_requires_unicode(dst)
               else open(src, 'rb'))
         with fh:
@@ -152,7 +124,7 @@ def _font_to_ps_type3(font_path, chars):
 """.format(font_name=font.postscript_name,
            inv_units_per_em=1 / font.units_per_EM,
            bbox=" ".join(map(str, font.bbox)),
-           encoding=" ".join("/{}".format(font.get_glyph_name(glyph_id))
+           encoding=" ".join(f"/{font.get_glyph_name(glyph_id)}"
                              for glyph_id in glyph_ids),
            num_glyphs=len(glyph_ids) + 1)
     postamble = """
@@ -176,7 +148,7 @@ FontName currentdict end definefont pop
 
     entries = []
     for glyph_id in glyph_ids:
-        g = font.load_glyph(glyph_id, LOAD_NO_SCALE)
+        g = font.load_glyph(glyph_id, LoadFlags.NO_SCALE)
         v, c = font.get_path()
         entries.append(
             "/%(name)s{%(bbox)s sc\n" % {
@@ -214,28 +186,201 @@ def _font_to_ps_type42(font_path, chars, fh):
     subset_str = ''.join(chr(c) for c in chars)
     _log.debug("SUBSET %s characters: %s", font_path, subset_str)
     try:
-        fontdata = _backend_pdf_ps.get_glyphs_subset(font_path, subset_str)
-        _log.debug("SUBSET %s %d -> %d", font_path, os.stat(font_path).st_size,
-                   fontdata.getbuffer().nbytes)
-
-        # Give ttconv a subsetted font along with updated glyph_ids.
-        font = FT2Font(fontdata)
-        glyph_ids = [font.get_char_index(c) for c in chars]
-        with TemporaryDirectory() as tmpdir:
-            tmpfile = os.path.join(tmpdir, "tmp.ttf")
-
-            with open(tmpfile, 'wb') as tmp:
-                tmp.write(fontdata.getvalue())
-
-            # TODO: allow convert_ttf_to_ps to input file objects (BytesIO)
-            convert_ttf_to_ps(os.fsencode(tmpfile), fh, 42, glyph_ids)
+        kw = {}
+        # fix this once we support loading more fonts from a collection
+        # https://github.com/matplotlib/matplotlib/issues/3135#issuecomment-571085541
+        if font_path.endswith('.ttc'):
+            kw['fontNumber'] = 0
+        with (fontTools.ttLib.TTFont(font_path, **kw) as font,
+              _backend_pdf_ps.get_glyphs_subset(font_path, subset_str) as subset):
+            fontdata = _backend_pdf_ps.font_as_file(subset).getvalue()
+            _log.debug(
+                "SUBSET %s %d -> %d", font_path, os.stat(font_path).st_size,
+                len(fontdata)
+            )
+            fh.write(_serialize_type42(font, subset, fontdata))
     except RuntimeError:
         _log.warning(
-            "The PostScript backend does not currently "
-            "support the selected font.")
+            "The PostScript backend does not currently support the selected font (%s).",
+            font_path)
         raise
 
 
+def _serialize_type42(font, subset, fontdata):
+    """
+    Output a PostScript Type-42 format representation of font
+
+    Parameters
+    ----------
+    font : fontTools.ttLib.ttFont.TTFont
+        The original font object
+    subset : fontTools.ttLib.ttFont.TTFont
+        The subset font object
+    fontdata : bytes
+        The raw font data in TTF format
+
+    Returns
+    -------
+    str
+        The Type-42 formatted font
+    """
+    version, breakpoints = _version_and_breakpoints(font.get('loca'), fontdata)
+    post = font['post']
+    name = font['name']
+    chars = _generate_charstrings(subset)
+    sfnts = _generate_sfnts(fontdata, subset, breakpoints)
+    return textwrap.dedent(f"""
+        %%!PS-TrueTypeFont-{version[0]}.{version[1]}-{font['head'].fontRevision:.7f}
+        10 dict begin
+        /FontType 42 def
+        /FontMatrix [1 0 0 1 0 0] def
+        /FontName /{name.getDebugName(6)} def
+        /FontInfo 7 dict dup begin
+        /FullName ({name.getDebugName(4)}) def
+        /FamilyName ({name.getDebugName(1)}) def
+        /Version ({name.getDebugName(5)}) def
+        /ItalicAngle {post.italicAngle} def
+        /isFixedPitch {'true' if post.isFixedPitch else 'false'} def
+        /UnderlinePosition {post.underlinePosition} def
+        /UnderlineThickness {post.underlineThickness} def
+        end readonly def
+        /Encoding StandardEncoding def
+        /FontBBox [{_nums_to_str(*_bounds(font))}] def
+        /PaintType 0 def
+        /CIDMap 0 def
+        {chars}
+        {sfnts}
+        FontName currentdict end definefont pop
+        """)
+
+
+def _version_and_breakpoints(loca, fontdata):
+    """
+    Read the version number of the font and determine sfnts breakpoints.
+
+    When a TrueType font file is written as a Type 42 font, it has to be
+    broken into substrings of at most 65535 bytes. These substrings must
+    begin at font table boundaries or glyph boundaries in the glyf table.
+    This function determines all possible breakpoints and it is the caller's
+    responsibility to do the splitting.
+
+    Helper function for _font_to_ps_type42.
+
+    Parameters
+    ----------
+    loca : fontTools.ttLib._l_o_c_a.table__l_o_c_a or None
+        The loca table of the font if available
+    fontdata : bytes
+        The raw data of the font
+
+    Returns
+    -------
+    version : tuple[int, int]
+        A 2-tuple of the major version number and minor version number.
+    breakpoints : list[int]
+        The breakpoints is a sorted list of offsets into fontdata; if loca is not
+        available, just the table boundaries.
+    """
+    v1, v2, numTables = struct.unpack('>3h', fontdata[:6])
+    version = (v1, v2)
+
+    tables = {}
+    for i in range(numTables):
+        tag, _, offset, _ = struct.unpack('>4sIII', fontdata[12 + i*16:12 + (i+1)*16])
+        tables[tag.decode('ascii')] = offset
+    if loca is not None:
+        glyf_breakpoints = {tables['glyf'] + offset for offset in loca.locations[:-1]}
+    else:
+        glyf_breakpoints = set()
+    breakpoints = sorted({*tables.values(), *glyf_breakpoints, len(fontdata)})
+
+    return version, breakpoints
+
+
+def _bounds(font):
+    """
+    Compute the font bounding box, as if all glyphs were written
+    at the same start position.
+
+    Helper function for _font_to_ps_type42.
+
+    Parameters
+    ----------
+    font : fontTools.ttLib.ttFont.TTFont
+        The font
+
+    Returns
+    -------
+    tuple
+        (xMin, yMin, xMax, yMax) of the combined bounding box
+        of all the glyphs in the font
+    """
+    gs = font.getGlyphSet(False)
+    pen = fontTools.pens.boundsPen.BoundsPen(gs)
+    for name in gs.keys():
+        gs[name].draw(pen)
+    return pen.bounds or (0, 0, 0, 0)
+
+
+def _generate_charstrings(font):
+    """
+    Transform font glyphs into CharStrings
+
+    Helper function for _font_to_ps_type42.
+
+    Parameters
+    ----------
+    font : fontTools.ttLib.ttFont.TTFont
+        The font
+
+    Returns
+    -------
+    str
+        A definition of the CharStrings dictionary in PostScript
+    """
+    go = font.getGlyphOrder()
+    s = f'/CharStrings {len(go)} dict dup begin\n'
+    for i, name in enumerate(go):
+        s += f'/{name} {i} def\n'
+    s += 'end readonly def'
+    return s
+
+
+def _generate_sfnts(fontdata, font, breakpoints):
+    """
+    Transform font data into PostScript sfnts format.
+
+    Helper function for _font_to_ps_type42.
+
+    Parameters
+    ----------
+    fontdata : bytes
+        The raw data of the font
+    font : fontTools.ttLib.ttFont.TTFont
+        The fontTools font object
+    breakpoints : list
+        Sorted offsets of possible breakpoints
+
+    Returns
+    -------
+    str
+        The sfnts array for the font definition, consisting
+        of hex-encoded strings in PostScript format
+    """
+    s = '/sfnts['
+    pos = 0
+    while pos < len(fontdata):
+        i = bisect.bisect_left(breakpoints, pos + 65534)
+        newpos = breakpoints[i-1]
+        if newpos <= pos:
+            # have to accept a larger string
+            newpos = breakpoints[-1]
+        s += f'<{fontdata[pos:newpos].hex()}00>'  # Always NUL terminate.
+        pos = newpos
+    s += ']def'
+    return '\n'.join(s[i:i+100] for i in range(0, len(s), 100))
+
+
 def _log_if_debug_on(meth):
     """
     Wrap `RendererPS` method *meth* to emit a PS comment with the method name,
@@ -284,7 +429,7 @@ class RendererPS(_backend_pdf_ps.RendererPDFPSBase):
         self._path_collection_id = 0
 
         self._character_tracker = _backend_pdf_ps.CharacterTracker()
-        self._logwarn_once = functools.lru_cache(None)(_log.warning)
+        self._logwarn_once = functools.cache(_log.warning)
 
     def _is_transparent(self, rgb_or_rgba):
         if rgb_or_rgba is None:
@@ -302,16 +447,16 @@ class RendererPS(_backend_pdf_ps.RendererPDFPSBase):
 
     def set_color(self, r, g, b, store=True):
         if (r, g, b) != self.color:
-            self._pswriter.write(f"{r:1.3f} setgray\n"
+            self._pswriter.write(f"{_nums_to_str(r)} setgray\n"
                                  if r == g == b else
-                                 f"{r:1.3f} {g:1.3f} {b:1.3f} setrgbcolor\n")
+                                 f"{_nums_to_str(r, g, b)} setrgbcolor\n")
             if store:
                 self.color = (r, g, b)
 
     def set_linewidth(self, linewidth, store=True):
         linewidth = float(linewidth)
         if linewidth != self.linewidth:
-            self._pswriter.write("%1.3f setlinewidth\n" % linewidth)
+            self._pswriter.write(f"{_nums_to_str(linewidth)} setlinewidth\n")
             if store:
                 self.linewidth = linewidth
 
@@ -347,8 +492,7 @@ class RendererPS(_backend_pdf_ps.RendererPDFPSBase):
             if np.array_equal(seq, oldseq) and oldo == offset:
                 return
 
-        self._pswriter.write(f"[{_nums_to_str(*seq)}]"
-                             f" {_nums_to_str(offset)} setdash\n"
+        self._pswriter.write(f"[{_nums_to_str(*seq)}] {_nums_to_str(offset)} setdash\n"
                              if seq is not None and len(seq) else
                              "[] 0 setdash\n")
         if store:
@@ -361,12 +505,11 @@ class RendererPS(_backend_pdf_ps.RendererPDFPSBase):
                 self.fontname = fontname
                 self.fontsize = fontsize
 
-    def create_hatch(self, hatch):
+    def create_hatch(self, hatch, linewidth):
         sidelen = 72
         if hatch in self._hatches:
             return self._hatches[hatch]
         name = 'H%d' % len(self._hatches)
-        linewidth = mpl.rcParams['hatch.linewidth']
         pageheight = self.height * 72
         self._pswriter.write(f"""\
   << /PatternType 1
@@ -379,8 +522,7 @@ class RendererPS(_backend_pdf_ps.RendererPDFPSBase):
      /PaintProc {{
         pop
         {linewidth:g} setlinewidth
-{self._convert_path(
-    Path.hatch(hatch), Affine2D().scale(sidelen), simplify=False)}
+{self._convert_path(Path.hatch(hatch), Affine2D().scale(sidelen), simplify=False)}
         gsave
         fill
         grestore
@@ -416,7 +558,7 @@ class RendererPS(_backend_pdf_ps.RendererPDFPSBase):
         clip = []
         rect = gc.get_clip_rectangle()
         if rect is not None:
-            clip.append("%s clipbox\n" % _nums_to_str(*rect.size, *rect.p0))
+            clip.append(f"{_nums_to_str(*rect.p0, *rect.size)} rectclip\n")
         path, trf = gc.get_clip_path()
         if path is not None:
             key = (path, id(trf))
@@ -483,9 +625,9 @@ grestore
         ps_color = (
             None
             if self._is_transparent(rgbFace)
-            else '%1.3f setgray' % rgbFace[0]
+            else f'{_nums_to_str(rgbFace[0])} setgray'
             if rgbFace[0] == rgbFace[1] == rgbFace[2]
-            else '%1.3f %1.3f %1.3f setrgbcolor' % rgbFace[:3])
+            else f'{_nums_to_str(*rgbFace[:3])} setrgbcolor')
 
         # construct the generic marker command:
 
@@ -523,7 +665,7 @@ grestore
                 simplify=False):
             if len(vertices):
                 x, y = vertices[-2:]
-                ps_cmd.append("%g %g o" % (x, y))
+                ps_cmd.append(f"{x:g} {y:g} o")
 
         ps = '\n'.join(ps_cmd)
         self._draw_ps(ps, gc, rgbFace, fill=False, stroke=False)
@@ -568,7 +710,7 @@ translate
                 gc, path_codes, offsets, offset_trans,
                 facecolors, edgecolors, linewidths, linestyles,
                 antialiaseds, urls, offset_position):
-            ps = "%g %g %s" % (xo, yo, path_id)
+            ps = f"{xo:g} {yo:g} {path_id}"
             self._draw_ps(ps, gc0, rgbFace)
 
         self._path_collection_id += 1
@@ -591,15 +733,17 @@ translate
         w, h, bl = self.get_text_width_height_descent(s, prop, ismath="TeX")
         fontsize = prop.get_size_in_points()
         thetext = 'psmarker%d' % self.textcnt
-        color = '%1.3f,%1.3f,%1.3f' % gc.get_rgb()[:3]
+        color = _nums_to_str(*gc.get_rgb()[:3], sep=',')
         fontcmd = {'sans-serif': r'{\sffamily %s}',
                    'monospace': r'{\ttfamily %s}'}.get(
                        mpl.rcParams['font.family'][0], r'{\rmfamily %s}')
         s = fontcmd % s
         tex = r'\color[rgb]{%s} %s' % (color, s)
 
-        # Stick to the bottom alignment.
-        pos = _nums_to_str(x, y-bl)
+        # Stick to bottom-left alignment, so subtract descent from the text-normal
+        # direction since text is normally positioned by its baseline.
+        rangle = np.radians(angle + 90)
+        pos = _nums_to_str(x - bl * np.cos(rangle), y - bl * np.sin(rangle))
         self.psfrag.append(
             r'\psfrag{%s}[bl][bl][1][%f]{\fontsize{%f}{%f}%s}' % (
                 thetext, angle, fontsize, fontsize*1.25, tex))
@@ -626,13 +770,15 @@ grestore
         if ismath:
             return self.draw_mathtext(gc, x, y, s, prop, angle)
 
+        stream = []  # list of (ps_name, x, char_name)
+
         if mpl.rcParams['ps.useafm']:
             font = self._get_font_afm(prop)
+            ps_name = (font.postscript_name.encode("ascii", "replace")
+                        .decode("ascii"))
             scale = 0.001 * prop.get_size_in_points()
-            stream = []
             thisx = 0
             last_name = None  # kerns returns 0 for None.
-            xs_names = []
             for c in s:
                 name = uni2type1.get(ord(c), f"uni{ord(c):04X}")
                 try:
@@ -643,38 +789,24 @@ grestore
                 kern = font.get_kern_dist_from_name(last_name, name)
                 last_name = name
                 thisx += kern * scale
-                xs_names.append((thisx, name))
+                stream.append((ps_name, thisx, name))
                 thisx += width * scale
-            ps_name = (font.postscript_name
-                       .encode("ascii", "replace").decode("ascii"))
-            stream.append((ps_name, xs_names))
 
         else:
             font = self._get_font_ttf(prop)
             self._character_tracker.track(font, s)
-            stream = []
-            prev_font = curr_stream = None
             for item in _text_helpers.layout(s, font):
                 ps_name = (item.ft_object.postscript_name
                            .encode("ascii", "replace").decode("ascii"))
-                if item.ft_object is not prev_font:
-                    if curr_stream:
-                        stream.append(curr_stream)
-                    prev_font = item.ft_object
-                    curr_stream = [ps_name, []]
-                curr_stream[1].append(
-                    (item.x, item.ft_object.get_glyph_name(item.glyph_idx))
-                )
-            # append the last entry if exists
-            if curr_stream:
-                stream.append(curr_stream)
-
+                glyph_name = item.ft_object.get_glyph_name(item.glyph_idx)
+                stream.append((ps_name, item.x, glyph_name))
         self.set_color(*gc.get_rgb())
 
-        for ps_name, xs_names in stream:
+        for ps_name, group in itertools. \
+                groupby(stream, lambda entry: entry[0]):
             self.set_font(ps_name, prop.get_size_in_points(), False)
             thetext = "\n".join(f"{x:g} 0 m /{name:s} glyphshow"
-                                for x, name in xs_names)
+                                for _, x, name in group)
             self._pswriter.write(f"""\
 gsave
 {self._get_clip_cmd(gc)}
@@ -711,14 +843,11 @@ grestore
             self._pswriter.write(f"{ox} {oy} {w} {h} rectfill\n")
         self._pswriter.write("grestore\n")
 
-    @_log_if_debug_on
-    def draw_gouraud_triangle(self, gc, points, colors, trans):
-        self.draw_gouraud_triangles(gc, points.reshape((1, 3, 2)),
-                                    colors.reshape((1, 3, 4)), trans)
-
     @_log_if_debug_on
     def draw_gouraud_triangles(self, gc, points, colors, trans):
         assert len(points) == len(colors)
+        if len(points) == 0:
+            return
         assert points.ndim == 3
         assert points.shape[1] == 3
         assert points.shape[2] == 2
@@ -803,10 +932,10 @@ grestore
                 write("grestore\n")
 
         if hatch:
-            hatch_name = self.create_hatch(hatch)
+            hatch_name = self.create_hatch(hatch, gc.get_hatch_linewidth())
             write("gsave\n")
-            write("%f %f %f " % gc.get_hatch_color()[:3])
-            write("%s setpattern fill grestore\n" % hatch_name)
+            write(_nums_to_str(*gc.get_hatch_color()[:3]))
+            write(f" {hatch_name} setpattern fill grestore\n")
 
         if stroke:
             write("stroke\n")
@@ -829,11 +958,10 @@ class FigureCanvasPS(FigureCanvasBase):
     def get_default_filetype(self):
         return 'ps'
 
-    @_api.delete_parameter("3.5", "args")
     def _print_ps(
-            self, fmt, outfile, *args,
+            self, fmt, outfile, *,
             metadata=None, papertype=None, orientation='portrait',
-            **kwargs):
+            bbox_inches_restore=None, **kwargs):
 
         dpi = self.figure.dpi
         self.figure.dpi = 72  # Override the dpi kwarg
@@ -849,8 +977,9 @@ class FigureCanvasPS(FigureCanvasBase):
         # See https://reproducible-builds.org/specs/source-date-epoch/
         source_date_epoch = os.getenv("SOURCE_DATE_EPOCH")
         dsc_comments["CreationDate"] = (
-            datetime.datetime.utcfromtimestamp(
-                int(source_date_epoch)).strftime("%a %b %d %H:%M:%S %Y")
+            datetime.datetime.fromtimestamp(
+                int(source_date_epoch),
+                datetime.timezone.utc).strftime("%a %b %d %H:%M:%S %Y")
             if source_date_epoch
             else time.ctime())
         dsc_comments = "\n".join(
@@ -859,7 +988,7 @@ class FigureCanvasPS(FigureCanvasBase):
         if papertype is None:
             papertype = mpl.rcParams['ps.papersize']
         papertype = papertype.lower()
-        _api.check_in_list(['auto', *papersize], papertype=papertype)
+        _api.check_in_list(['figure', *papersize], papertype=papertype)
 
         orientation = _api.check_getitem(
             _Orientation, orientation=orientation.lower())
@@ -868,7 +997,8 @@ class FigureCanvasPS(FigureCanvasBase):
                    if mpl.rcParams['text.usetex'] else
                    self._print_figure)
         printer(fmt, outfile, dpi=dpi, dsc_comments=dsc_comments,
-                orientation=orientation, papertype=papertype, **kwargs)
+                orientation=orientation, papertype=papertype,
+                bbox_inches_restore=bbox_inches_restore, **kwargs)
 
     def _print_figure(
             self, fmt, outfile, *,
@@ -878,7 +1008,7 @@ class FigureCanvasPS(FigureCanvasBase):
         Render the figure to a filesystem path or a file-like object.
 
         Parameters are as for `.print_figure`, except that *dsc_comments* is a
-        all string containing Document Structuring Convention comments,
+        string containing Document Structuring Convention comments,
         generated from the *metadata* parameter to `.print_figure`.
         """
         is_eps = fmt == 'eps'
@@ -888,19 +1018,11 @@ class FigureCanvasPS(FigureCanvasBase):
 
         # find the appropriate papertype
         width, height = self.figure.get_size_inches()
-        if papertype == 'auto':
-            papertype = _get_papertype(
-                *orientation.swap_if_landscape((width, height)))
-        paper_width, paper_height = orientation.swap_if_landscape(
-            papersize[papertype])
-
-        if mpl.rcParams['ps.usedistiller']:
-            # distillers improperly clip eps files if pagesize is too small
-            if width > paper_width or height > paper_height:
-                papertype = _get_papertype(
-                    *orientation.swap_if_landscape((width, height)))
-                paper_width, paper_height = orientation.swap_if_landscape(
-                    papersize[papertype])
+        if is_eps or papertype == 'figure':
+            paper_width, paper_height = width, height
+        else:
+            paper_width, paper_height = orientation.swap_if_landscape(
+                papersize[papertype])
 
         # center the figure on the paper
         xo = 72 * 0.5 * (paper_width - width)
@@ -932,23 +1054,24 @@ class FigureCanvasPS(FigureCanvasBase):
             if is_eps:
                 print("%!PS-Adobe-3.0 EPSF-3.0", file=fh)
             else:
-                print(f"%!PS-Adobe-3.0\n"
-                      f"%%DocumentPaperSizes: {papertype}\n"
-                      f"%%Pages: 1\n",
-                      end="", file=fh)
-            print(f"{dsc_comments}\n"
+                print("%!PS-Adobe-3.0", file=fh)
+                if papertype != 'figure':
+                    print(f"%%DocumentPaperSizes: {papertype}", file=fh)
+                print("%%Pages: 1", file=fh)
+            print(f"%%LanguageLevel: 3\n"
+                  f"{dsc_comments}\n"
                   f"%%Orientation: {orientation.name}\n"
-                  f"{get_bbox_header(bbox)[0]}\n"
+                  f"{_get_bbox_header(bbox)}\n"
                   f"%%EndComments\n",
                   end="", file=fh)
 
-            Ndict = len(psDefs)
+            Ndict = len(_psDefs)
             print("%%BeginProlog", file=fh)
             if not mpl.rcParams['ps.useafm']:
                 Ndict += len(ps_renderer._character_tracker.used)
             print("/mpldict %d dict def" % Ndict, file=fh)
             print("mpldict begin", file=fh)
-            print("\n".join(psDefs), file=fh)
+            print("\n".join(_psDefs), file=fh)
             if not mpl.rcParams['ps.useafm']:
                 for font_path, chars \
                         in ps_renderer._character_tracker.used.items():
@@ -973,8 +1096,7 @@ class FigureCanvasPS(FigureCanvasBase):
             print("%s translate" % _nums_to_str(xo, yo), file=fh)
             if rotation:
                 print("%d rotate" % rotation, file=fh)
-            print("%s clipbox" % _nums_to_str(width*72, height*72, 0, 0),
-                  file=fh)
+            print(f"0 0 {_nums_to_str(width*72, height*72)} rectclip", file=fh)
 
             # write the figure
             print(self._pswriter.getvalue(), file=fh)
@@ -1046,18 +1168,19 @@ class FigureCanvasPS(FigureCanvasBase):
             tmppath.write_text(
                 f"""\
 %!PS-Adobe-3.0 EPSF-3.0
+%%LanguageLevel: 3
 {dsc_comments}
-{get_bbox_header(bbox)[0]}
+{_get_bbox_header(bbox)}
 %%EndComments
 %%BeginProlog
-/mpldict {len(psDefs)} dict def
+/mpldict {len(_psDefs)} dict def
 mpldict begin
-{"".join(psDefs)}
+{"".join(_psDefs)}
 end
 %%EndProlog
 mpldict begin
 {_nums_to_str(xo, yo)} translate
-{_nums_to_str(width*72, height*72)} 0 0 clipbox
+0 0 {_nums_to_str(width*72, height*72)} rectclip
 {self._pswriter.getvalue()}
 end
 showpage
@@ -1071,12 +1194,10 @@ showpage
             # set the paper size to the figure size if is_eps. The
             # resulting ps file has the given size with correct bounding
             # box so that there is no need to call 'pstoeps'
-            if is_eps:
+            if is_eps or papertype == 'figure':
                 paper_width, paper_height = orientation.swap_if_landscape(
                     self.figure.get_size_inches())
             else:
-                if papertype == 'auto':
-                    papertype = _get_papertype(width, height)
                 paper_width, paper_height = papersize[papertype]
 
             psfrag_rotated = _convert_psfrags(
@@ -1103,13 +1224,6 @@ showpage
         return super().draw()
 
 
-@_api.deprecated("3.6")
-def convert_psfrags(tmpfile, psfrags, font_preamble, custom_preamble,
-                    paper_width, paper_height, orientation):
-    return _convert_psfrags(
-        pathlib.Path(tmpfile), psfrags, paper_width, paper_height, orientation)
-
-
 def _convert_psfrags(tmppath, psfrags, paper_width, paper_height, orientation):
     """
     When we want to use the LaTeX backend with postscript, we write PSFrag tags
@@ -1174,9 +1288,14 @@ def gs_distill(tmpfile, eps=False, ptype='letter', bbox=None, rotated=False):
     """
 
     if eps:
-        paper_option = "-dEPSCrop"
+        paper_option = ["-dEPSCrop"]
+    elif ptype == "figure":
+        # The bbox will have its lower-left corner at (0, 0), so upper-right
+        # corner corresponds with paper size.
+        paper_option = [f"-dDEVICEWIDTHPOINTS={bbox[2]}",
+                        f"-dDEVICEHEIGHTPOINTS={bbox[3]}"]
     else:
-        paper_option = "-sPAPERSIZE=%s" % ptype
+        paper_option = [f"-sPAPERSIZE={ptype}"]
 
     psfile = tmpfile + '.ps'
     dpi = mpl.rcParams['ps.distiller.res']
@@ -1184,7 +1303,7 @@ def gs_distill(tmpfile, eps=False, ptype='letter', bbox=None, rotated=False):
     cbook._check_and_log_subprocess(
         [mpl._get_executable_info("gs").executable,
          "-dBATCH", "-dNOPAUSE", "-r%d" % dpi, "-sDEVICE=ps2write",
-         paper_option, "-sOutputFile=%s" % psfile, tmpfile],
+         *paper_option, f"-sOutputFile={psfile}", tmpfile],
         _log)
 
     os.remove(tmpfile)
@@ -1195,7 +1314,7 @@ def gs_distill(tmpfile, eps=False, ptype='letter', bbox=None, rotated=False):
     # the original bbox can be restored during the pstoeps step.
 
     if eps:
-        # For some versions of gs, above steps result in an ps file where the
+        # For some versions of gs, above steps result in a ps file where the
         # original bbox is no more correct. Do not adjust bbox for now.
         pstoeps(tmpfile, bbox, rotated=rotated)
 
@@ -1210,6 +1329,16 @@ def xpdf_distill(tmpfile, eps=False, ptype='letter', bbox=None, rotated=False):
     mpl._get_executable_info("gs")  # Effectively checks for ps2pdf.
     mpl._get_executable_info("pdftops")
 
+    if eps:
+        paper_option = ["-dEPSCrop"]
+    elif ptype == "figure":
+        # The bbox will have its lower-left corner at (0, 0), so upper-right
+        # corner corresponds with paper size.
+        paper_option = [f"-dDEVICEWIDTHPOINTS#{bbox[2]}",
+                        f"-dDEVICEHEIGHTPOINTS#{bbox[3]}"]
+    else:
+        paper_option = [f"-sPAPERSIZE#{ptype}"]
+
     with TemporaryDirectory() as tmpdir:
         tmppdf = pathlib.Path(tmpdir, "tmp.pdf")
         tmpps = pathlib.Path(tmpdir, "tmp.ps")
@@ -1222,31 +1351,35 @@ def xpdf_distill(tmpfile, eps=False, ptype='letter', bbox=None, rotated=False):
              "-sAutoRotatePages#None",
              "-sGrayImageFilter#FlateEncode",
              "-sColorImageFilter#FlateEncode",
-             "-dEPSCrop" if eps else "-sPAPERSIZE#%s" % ptype,
+             *paper_option,
              tmpfile, tmppdf], _log)
         cbook._check_and_log_subprocess(
-            ["pdftops", "-paper", "match", "-level2", tmppdf, tmpps], _log)
+            ["pdftops", "-paper", "match", "-level3", tmppdf, tmpps], _log)
         shutil.move(tmpps, tmpfile)
     if eps:
         pstoeps(tmpfile)
 
 
+@_api.deprecated("3.9")
 def get_bbox_header(lbrt, rotated=False):
     """
     Return a postscript header string for the given bbox lbrt=(l, b, r, t).
     Optionally, return rotate command.
     """
+    return _get_bbox_header(lbrt), (_get_rotate_command(lbrt) if rotated else "")
 
+
+def _get_bbox_header(lbrt):
+    """Return a PostScript header string for bounding box *lbrt*=(l, b, r, t)."""
     l, b, r, t = lbrt
-    if rotated:
-        rotate = "%.2f %.2f translate\n90 rotate" % (l+r, 0)
-    else:
-        rotate = ""
-    bbox_info = '%%%%BoundingBox: %d %d %d %d' % (l, b, np.ceil(r), np.ceil(t))
-    hires_bbox_info = '%%%%HiResBoundingBox: %.6f %.6f %.6f %.6f' % (
-        l, b, r, t)
+    return (f"%%BoundingBox: {int(l)} {int(b)} {math.ceil(r)} {math.ceil(t)}\n"
+            f"%%HiResBoundingBox: {l:.6f} {b:.6f} {r:.6f} {t:.6f}")
 
-    return '\n'.join([bbox_info, hires_bbox_info]), rotate
+
+def _get_rotate_command(lbrt):
+    """Return a PostScript 90° rotation command for bounding box *lbrt*=(l, b, r, t)."""
+    l, b, r, t = lbrt
+    return f"{l+r:.2f} {0:.2f} translate\n90 rotate"
 
 
 def pstoeps(tmpfile, bbox=None, rotated=False):
@@ -1256,12 +1389,6 @@ def pstoeps(tmpfile, bbox=None, rotated=False):
     None, original bbox will be used.
     """
 
-    # if rotated==True, the output eps file need to be rotated
-    if bbox:
-        bbox_info, rotate = get_bbox_header(bbox, rotated=rotated)
-    else:
-        bbox_info, rotate = None, None
-
     epsfile = tmpfile + '.eps'
     with open(epsfile, 'wb') as epsh, open(tmpfile, 'rb') as tmph:
         write = epsh.write
@@ -1270,7 +1397,7 @@ def pstoeps(tmpfile, bbox=None, rotated=False):
             if line.startswith(b'%!PS'):
                 write(b"%!PS-Adobe-3.0 EPSF-3.0\n")
                 if bbox:
-                    write(bbox_info.encode('ascii') + b'\n')
+                    write(_get_bbox_header(bbox).encode('ascii') + b'\n')
             elif line.startswith(b'%%EndComments'):
                 write(line)
                 write(b'%%BeginProlog\n'
@@ -1282,8 +1409,8 @@ def pstoeps(tmpfile, bbox=None, rotated=False):
                       b'/setpagedevice {pop} def\n'
                       b'%%EndProlog\n'
                       b'%%Page 1 1\n')
-                if rotate:
-                    write(rotate.encode('ascii') + b'\n')
+                if rotated:  # The output eps file need to be rotated.
+                    write(_get_rotate_command(bbox).encode('ascii') + b'\n')
                 break
             elif bbox and line.startswith((b'%%Bound', b'%%HiResBound',
                                            b'%%DocumentMedia', b'%%Pages')):
@@ -1325,7 +1452,7 @@ FigureManagerPS = FigureManagerBase
 
 # The usage comments use the notation of the operator summary
 # in the PostScript Language reference manual.
-psDefs = [
+_psDefs = [
     # name proc  *_d*  -
     # Note that this cannot be bound to /d, because when embedding a Type3 font
     # we may want to define a "d" glyph using "/d{...} d" which would locally
@@ -1343,20 +1470,6 @@ psDefs = [
     "/cl { closepath } _d",
     # *ce*  -
     "/ce { closepath eofill } _d",
-    # w h x y  *box*  -
-    """/box {
-      m
-      1 index 0 r
-      0 exch r
-      neg 0 r
-      cl
-    } _d""",
-    # w h x y  *clipbox*  -
-    """/clipbox {
-      box
-      clip
-      newpath
-    } _d""",
     # wx wy llx lly urx ury  *setcachedevice*  -
     "/sc { setcachedevice } _d",
 ]
@@ -1364,4 +1477,5 @@ psDefs = [
 
 @_Backend.export
 class _BackendPS(_Backend):
+    backend_version = 'Level II'
     FigureCanvas = FigureCanvasPS
